﻿using Newtonsoft.Json.Linq;
using OBSBlur.Window;
using OBSWebsocketDotNet;
using OBSWebsocketDotNet.Communication;
using OBSWebsocketDotNet.Types;
using OBSWebsocketDotNet.Types.Events;

namespace OBSBlur.OBS;

public class Blur
{
    WindowManager windowManager = new ();
    private OBSWebsocket _websocket = new ();
    private string currentScene;
    private string[] enabledScenes;
    private string[] blurPrograms;
    private Dictionary<IntPtr, uint> windowHandleSceneItems = new();

    public Blur(string obsUrl, string obsPassword, string[] enabledScenes, string[] blurPrograms)
    {
        this.enabledScenes = enabledScenes;
        this.blurPrograms = blurPrograms;
        _websocket.CurrentProgramSceneChanged += WebsocketOnCurrentProgramSceneChanged;
        _websocket.Connected += WebsocketOnConnected;
        _websocket.Disconnected += WebsocketOnDisconnected;
        
        _websocket.ConnectAsync(obsUrl, obsPassword);
        
        windowManager.WindowsChanged += WindowManagerOnWindowsChanged;
        windowManager.WindowZOrderChanged += WindowManagerOnWindowZOrderChanged;
    }

    private void UpdateBlurs(WindowInfo[]? windowInfos = null, IntPtr[]? zOrder = null)
    {
        if (!enabledScenes.Contains(currentScene))
            return;
        if (!_websocket.IsConnected)
            return;
        windowInfos ??= windowManager.WindowInfos;
        zOrder ??= windowManager.WindowZOrder;

        bool maximixedWindowReached = false;
        
        foreach (IntPtr window in zOrder)
        {
            WindowInfo windowInfo = windowInfos.FirstOrDefault(w => w.WindowHandle == window);
            if(windowInfo is {WindowHandle: 0x0})//No WindowInfo found
                continue;
            
            if(GetBlurWindow(windowInfo))
                if(!maximixedWindowReached)
                    BlurWindow(windowInfo);
                else if(maximixedWindowReached) 
                    DeleteBlur(windowInfo);
            
            if (windowInfo.WindowCommands is ShowWindowCommands.Maximize or ShowWindowCommands.ShowMaximized)
                maximixedWindowReached = true;
        }
        
        foreach(IntPtr blurredWindow in windowHandleSceneItems.Keys.ToArray())
            if(!windowInfos.Any(w => w.WindowHandle == blurredWindow))
                DeleteBlur(blurredWindow);
    }

    private bool GetBlurWindow(WindowInfo windowInfo)
    {
        foreach(string program in blurPrograms)
            if (windowInfo.ProcessInfo.ProcessName.Contains(program))
                return true;
        return false;
    }

    private void BlurWindow(WindowInfo windowInfo)
    {
        if(windowHandleSceneItems.ContainsKey(windowInfo.WindowHandle))
            MoveBlur(windowInfo);
        else
            AddBlur(windowInfo);
    }

    private void AddBlur(WindowInfo windowInfo)
    {
        Dictionary<string, object> request = new()
        {
            {"sceneName", currentScene},
            {"sceneItemId", GetBlurSource()}
        };
        try
        {
            JObject response = _websocket.SendRequest("DuplicateSceneItem", JObject.FromObject(request));
            windowHandleSceneItems.Add(windowInfo.WindowHandle, response["sceneItemId"]!.Value<uint>());
            MoveBlur(windowInfo);
        }
        catch (ErrorResponseException e)
        {
            Console.WriteLine("No Scene Item with name 'Blur' found.");
            throw;
        }
    }

    private void MoveBlur(WindowInfo windowInfo)
    {
        if (!windowHandleSceneItems.ContainsKey(windowInfo.WindowHandle))
            return;

        SceneItemTransformInfo info = new()
        {
            X = windowInfo.WindowRectangle.X,
            Y = windowInfo.WindowRectangle.Y,
            BoundsHeight = windowInfo.WindowRectangle.Height,
            BoundsWidth = windowInfo.WindowRectangle.Width,
            BoundsType = SceneItemBoundsType.OBS_BOUNDS_STRETCH,
            Alignnment = 5
        };
        
        Dictionary<string, object> request = new()
        {
            {"sceneName", currentScene},
            {"sceneItemId", windowHandleSceneItems[windowInfo.WindowHandle]},
            {"sceneItemTransform", info}
        };
        _websocket.SendRequest("SetSceneItemTransform", JObject.FromObject(request));
        
        
        Dictionary<string, object> request2 = new()
        {
            {"sceneName", currentScene},
            {"sceneItemId", windowHandleSceneItems[windowInfo.WindowHandle]},
            {"sceneItemEnabled", true}
        };
        _websocket.SendRequest("SetSceneItemEnabled", JObject.FromObject(request2));
    }

    private void DeleteBlur(IntPtr windowHandle)
    {
        if (!windowHandleSceneItems.ContainsKey(windowHandle))
            return;
        Dictionary<string, object> request = new()
        {
            {"sceneName", currentScene},
            {"sceneItemId", windowHandleSceneItems[windowHandle]}
        };
        _websocket.SendRequest("RemoveSceneItem", JObject.FromObject(request));
        windowHandleSceneItems.Remove(windowHandle);
    }

    private void DeleteBlur(WindowInfo windowInfo)
    {
        DeleteBlur(windowInfo.WindowHandle);
    }

    private uint GetBlurSource()
    {
        Dictionary<string, string> request = new()
        {
            {"sceneName", currentScene},
            {"sourceName", "Blur"}
        };
        try
        {
            JObject response = _websocket.SendRequest("GetSceneItemId", JObject.FromObject(request));
            return response["sceneItemId"]!.Value<uint>();
        }
        catch (ErrorResponseException e)
        {
            Console.WriteLine("No Scene Item with name 'Blur' found.");
            throw;
        }
    }

    private void WindowManagerOnWindowZOrderChanged(IntPtr[] neworder)
    {
        
        uint i = 0;
        string prnt = "";
        foreach (IntPtr windowHandle in neworder)
        {
            WindowInfo windowInfo = windowManager.WindowInfos.FirstOrDefault(w => w.WindowHandle == windowHandle);
            if (windowInfo.WindowHandle is 0x0)
                continue;
            prnt += $"{++i,3} | {windowInfo}\n";
        }

        Console.WriteLine($"Z-order changed {DateTime.UtcNow:O}");
        Console.WriteLine($"{"Z",3} | {"hWnd",8} | {"PID",8} | {"Name",17} | {"Window Title",-17} | {"Vis",-8} | {"State",-13} | Rectangle");
        Console.WriteLine(prnt);
        UpdateBlurs(zOrder: neworder);
    }

    private void WindowManagerOnWindowsChanged(WindowInfo[] before, WindowInfo[] after)
    {
        string prnt = "";
        foreach (WindowInfo windowInfo in after)
            prnt += $"{windowInfo}\n";
        Console.WriteLine($"Windows changed {DateTime.UtcNow:O}");
        Console.WriteLine(prnt);
        UpdateBlurs(windowInfos: after);
    }

    private void WebsocketOnDisconnected(object? sender, ObsDisconnectionInfo e)
    {
        Console.WriteLine($"Obs Disconnected {DateTime.UtcNow:O}");
    }

    private void WebsocketOnConnected(object? sender, EventArgs e)
    {
        currentScene = _websocket.GetCurrentProgramScene();
        Console.WriteLine($"Obs Connected. Current Scene {currentScene} {DateTime.UtcNow:O}");
        UpdateBlurs();
    }

    private void WebsocketOnCurrentProgramSceneChanged(object? sender, ProgramSceneChangedEventArgs e)
    {
        currentScene = e.SceneName;
        Console.WriteLine($"Obs Scene Changed -> {currentScene} {DateTime.UtcNow:O}");
        UpdateBlurs();
    }
    
    
}